(* Some tools *)

let memo f =
  let h = Hashtbl.create 101 in (* on choisit un nombre premier adéquate *)
    fun x -> try Hashtbl.find h x with
	Not_found ->
	  let y = f x in
	    Hashtbl.add h x y; y

let memo_rec f =
  let h = Hashtbl.create 101 in
  let rec aux x =
    try Hashtbl.find h x with
	      Not_found ->
	        let y = f aux x in
	        Hashtbl.add h x y; y
  in
  aux

let is_prime =
  memo (
    fun n ->
      let rec aux = function
	      | 0|1 -> true
	      | x -> n mod x <> 0 && aux (x-1)
      in aux (n/2)
  )

let next_prime =
  memo_rec (
    fun f n ->
      if is_prime n then n
      else f (n+1)
  )

module StringSet = Set.Make(String)

let get_succ x s =
  try Hashtbl.find s x with Not_found -> StringSet.empty

let extend r a b =
  Hashtbl.replace r a (StringSet.add b (get_succ a r))

let make l =
  let r = Hashtbl.create (next_prime (List.length l)) in
  List.iter (fun (x,y) -> extend r x y) l;
  r

let print_rel r =
  Hashtbl.iter (
    fun a b ->
      Printf.printf "%s : [ " a;
      StringSet.iter (Printf.printf "%s; ") b;
      Printf.printf "]\n"
  ) r

let count =
  let c = ref 0 in
  fun n -> c := !c + n; !c

let is_in a b r =
  let marked = ref (StringSet.add a StringSet.empty) in
  let rec aux x =
    (StringSet.mem b (get_succ x r))
    || StringSet.exists
      (fun z ->
	      (not (StringSet.mem z !marked))
	      && (marked := StringSet.add z !marked;
	          aux z)
      ) (get_succ x r)
  in aux a
  && (
    let c = StringSet.cardinal !marked - 1 in
	  Printf.printf "tested: %d (%d total)\n" c (count c);
	  true
  )

let is_in_memo a b r =
  let marked = ref (StringSet.add a StringSet.empty) in
  let rec aux x =
    (StringSet.mem b (get_succ x r) && (extend r a b; true))
    || StringSet.exists
      (fun z ->
	      (not (StringSet.mem z !marked))
	      && (marked := StringSet.add z !marked;
	          extend r a z;
	          aux z && (extend r z b; true))
      ) (get_succ x r)
  in aux a
  && (
    let c = StringSet.cardinal !marked - 1 in
	  Printf.printf "tested: %d (%d total)\n" c (count c);
	  true
  )

(* Pour ceux que ca interesse voici une version en largeur avec memoization *)
exception Key_found

let is_in_larg a b s =
  let q = Queue.create () in
  let marked = ref StringSet.empty in
  begin
    Queue.push a q;
    marked := StringSet.add a !marked;
    try
	    while not (Queue.is_empty q) do
	      let x = Queue.take q in
	      let succ = get_succ x s in
	      ignore (StringSet.mem b succ && (extend s a b; raise Key_found));
	      StringSet.iter
	        (fun y ->
		        if not (StringSet.mem y !marked) then
		          begin
		            extend s a y;
		            Queue.push y q;
		            marked := StringSet.add y !marked;
		          end
	        ) succ;
	    done;
	    false
    with
	      Key_found ->
	        Printf.printf "tested: %d\n" (StringSet.cardinal !marked);
	        true
  end

let read_pair cin =
  let l = input_line cin in
  let p = String.index l ' ' in
  (String.sub l 0 p, String.sub l (p+1) ((String.length l) - (p+1)))

let rec from_file cin =
  try
    let p = read_pair cin in p :: from_file cin
  with _ -> []

let main () =
  if (Array.length Sys.argv < 2) then invalid_arg "Il faut un nom de fichier en entrée!";
  let r = make (from_file (open_in Sys.argv.(1))) in
  let test =
    if Array.length Sys.argv > 2 && Sys.argv.(2) <> "memo" then is_in
    else is_in_memo
  in
  let nb = ref 0 in
  begin
    print_rel r;
    Printf.printf "-- EVAL --\n\n%!";
    (try
	     while true do
	       let (a,b) = read_pair stdin in
	       if test a b r then
	         Printf.printf "(%s,%s) est dans r\n\n%!" a b
	       else
	         Printf.printf "(%s,%s) n'est pas dans r\n\n%!" a b;
	       incr nb;
	     done
     with End_of_file -> ());
    Printf.printf "-- END EVAL --\n%!";
    Printf.printf "tests: %d\nValeurs testées: %d\n\n%!" !nb (count 0);
    print_rel r;
    exit 0;
  end

let _ = main ()
